import math
from functools import cached_property
from collections import defaultdict
from typing import Callable, Tuple

from flippy.tools import isclose
from flippy.distributions import Categorical, RandomNumberGenerator, \
    Dirichlet, Uniform, Beta
from flippy.distributions.random import default_rng
from flippy.distributions.base import Distribution, FiniteDistribution
from flippy.distributions.support import Interval, Simplex
from flippy.core import ReturnState, SampleState, ObserveState, ProgramState
from flippy.interpreter import CPSInterpreter
from flippy.inference.inference import InferenceAlgorithm
from flippy.types import Element

from flippy.inference.mcmc.trace import Trace
from flippy.inference.mcmc.diagnostics import MCMCDiagnostics, MCMCDiagnosticsEntry

from flippy.types import VariableName, SampleValue

ProposalKernel = Callable[['SampleValue', SampleState], Distribution['SampleValue']]

class MetropolisHastings(InferenceAlgorithm[Element]):
    """
    Markov Chain Monte Carlo (MCMC) inference using the Metropolis-Hastings algorithm.
    - `function`: The function to be executed
    - `samples`: Number of samples to draw from the posterior
    - `burn_in`: Number of initial samples to discard (default: 0)
    - `thinning`: Interval between saved samples (default: 1, i.e., save every sample)
    - `seed`: Optional random seed for reproducibility
    - `use_drift_kernels`: Whether to use drift kernels for continuous variables (default: True)
    - `uniform_drift_kernel_width`: Width of the uniform drift kernel for bounded continuous variables
    - `simplex_proposal_kernel_alpha`: Concentration parameter for the Dirichlet proposal kernel for simplex variables
    - `custom_proposal_kernels`: Optional function that takes a variable name and returns a
    proposal kernel function for that variable
    - `custom_initial_trace_kernel`: Optional function that takes a variable name and returns an initial
    value for that variable
    """
    # this is finite in case it is impossible to initialize a trace
    max_initial_trace_attempts = 1000
    def __init__(
        self,
        function : Callable,
        samples : int,
        burn_in : int = 0,
        thinning : int = 1,
        seed : int = None,
        use_drift_kernels : bool = True,
        uniform_drift_kernel_width : float = 1,
        simplex_proposal_kernel_alpha : float = 10,
        custom_proposal_kernels : Callable[['VariableName'], ProposalKernel] = None,
        custom_initial_trace_kernel : Callable[['VariableName'], 'SampleValue'] = None,
    ):
        self.function = function
        self.samples = samples
        self.burn_in = burn_in
        self.thinning = thinning
        self.save_diagnostics = False
        self.seed = seed
        self.use_drift_kernels = use_drift_kernels
        self.uniform_drift_kernel_width = uniform_drift_kernel_width
        self.simplex_proposal_kernel_alpha = simplex_proposal_kernel_alpha
        self.custom_proposal_kernels = custom_proposal_kernels
        self.custom_initial_trace_kernel = custom_initial_trace_kernel

    @property
    def is_cachable(self):
        return self.seed is not None

    def run(self, *args, **kws) -> Distribution[Element]:
        self.save_diagnostics = False
        dist, _ = self._run(*args, **kws)
        return dist

    def run_with_diagnostics(self, *args, **kws) -> Tuple[Distribution, MCMCDiagnostics]:
        self.save_diagnostics = True
        dist, diagnostics = self._run(*args, **kws)
        return dist, diagnostics

    def _run(self, *args, **kws):
        init_ps = self.initial_program_state
        init_ps = init_ps.step(*args, **kws)
        rng = RandomNumberGenerator(self.seed)
        initial_trace = self.generate_initial_trace(init_ps, rng)
        assert initial_trace.total_score > float('-inf')
        return self.run_from_initial_trace(initial_trace, rng=rng)

    @cached_property
    def initial_program_state(self):
        return CPSInterpreter().initial_program_state(self.function)

    def run_from_initial_trace(
        self,
        initial_trace : Trace,
        rng : RandomNumberGenerator = default_rng
    ):
        diagnostics = MCMCDiagnostics()
        return_counts = defaultdict(int)
        old_trace = initial_trace
        iterator = range(self.burn_in + self.samples*self.thinning)
        for i in iterator:
            target_site_name, old_site_score = \
                self.choose_target_site(
                    trace=old_trace,
                    rng=rng
                )
            new_trace, new_proposal_score = \
                self.choose_new_trace(
                    old_trace=old_trace,
                    target_site_name=target_site_name,
                    rng=rng
                )
            _target_site_name, new_site_score = \
                self.choose_target_site(
                    trace=new_trace,
                    target_site_name=target_site_name,
                    rng=rng
                )
            assert target_site_name == _target_site_name
            _, old_proposal_score = \
                self.choose_new_trace(
                    old_trace=new_trace,
                    target_site_name=target_site_name,
                    new_trace=old_trace,
                    rng=rng
                )
            log_acceptance_ratio = self.calc_log_acceptance_ratio(
                new_score=new_trace.total_score,
                old_score=old_trace.total_score,
                new_site_score=new_site_score,
                old_proposal_score=old_proposal_score,
                old_site_score=old_site_score,
                new_proposal_score=new_proposal_score
            )
            log_acceptance_threshold = math.log(rng.random())
            accept = log_acceptance_ratio > log_acceptance_threshold

            if accept:
                old_trace = new_trace
            save_sample = (i >= self.burn_in) and (((i - self.burn_in) % self.thinning) == 0)
            if save_sample:
                return_counts[old_trace.return_value] += 1
            if self.save_diagnostics:
                diagnostics.append(MCMCDiagnosticsEntry(
                    old_trace=old_trace,
                    new_trace=new_trace,
                    sampled_trace=new_trace if accept else old_trace,
                    accept=accept,
                    log_acceptance_threshold=log_acceptance_threshold,
                    log_acceptance_ratio=log_acceptance_ratio,
                    save_sample=save_sample,
                    auxiliary_vars=target_site_name,
                ))
        assert sum(return_counts.values()) == self.samples, (sum(return_counts.values()), self.samples)
        return (
            Categorical.from_dict({e: c/self.samples for e, c in return_counts.items()}),
            diagnostics
        )

    def calc_log_acceptance_ratio(
        self,
        new_score,
        old_score,
        new_site_score,
        old_proposal_score,
        old_site_score,
        new_proposal_score,
    ):
        log_acceptance_num = new_score + new_site_score + old_proposal_score
        log_acceptance_den = old_score + old_site_score + new_proposal_score
        if log_acceptance_num == float('-inf'):
            return float('-inf')
        else:
            log_acceptance_ratio = log_acceptance_num - log_acceptance_den
            assert not math.isnan(log_acceptance_ratio)
            return log_acceptance_ratio

    def generate_initial_trace(
        self,
        initial_program_state : ProgramState,
        rng : RandomNumberGenerator = default_rng
    ) -> Trace:
        def sample_site_callback(ps : SampleState) -> 'SampleValue':
            assert not ps.fit, f"MetropolisHastings doesn't support Distribution.fit: {ps.name}"
            if (
                ps.initial_value is not None and \
                ps.initial_value in ps.distribution.support
            ):
                value = ps.initial_value
            elif self.custom_initial_trace_kernel is not None:
                value = self.custom_initial_trace_kernel(ps.name)
                if value is None:
                    value = ps.distribution.sample(rng=rng)
            else:
                value = ps.distribution.sample(rng=rng)
            return value

        iterator = range(self.max_initial_trace_attempts)
        for i in iterator:
            trace = Trace.run_from(
                ps=initial_program_state,
                old_trace=None,
                sample_site_callback=sample_site_callback,
                observe_site_callback=lambda ps : ps.value
            )
            if trace.total_score > float('-inf'):
                return trace
        raise RuntimeError(f'Could not generate initial trace after {self.max_initial_trace_attempts} attempts')

    def choose_target_site(
        self,
        trace : Trace,
        target_site_name : 'VariableName' = None,
        rng : RandomNumberGenerator = default_rng,
    ) -> Tuple['VariableName', float]:
        sample_sites = [e.name for e in trace.entries() if e.is_sample]
        if target_site_name is None:
            return rng.choice(sample_sites), math.log(1/len(sample_sites))
        else:
            log_prob = math.log(1/len(sample_sites)) if target_site_name in sample_sites else float('-inf')
            return target_site_name, log_prob

    def choose_new_trace(
        self,
        old_trace : Trace,
        target_site_name : 'VariableName',
        new_trace : Trace = None,
        rng : RandomNumberGenerator = default_rng,
    ) -> Tuple[Trace, float]:
        if new_trace is None:
            new_trace = self.sample_new_trace(
                old_trace=old_trace,
                target_site_name=target_site_name,
                rng=rng
            )
        log_prob = self.calc_new_trace_log_probability(
            old_trace=old_trace,
            target_site_name=target_site_name,
            new_trace=new_trace
        )
        return new_trace, log_prob

    def sample_new_trace(
        self,
        old_trace : Trace,
        target_site_name : 'VariableName',
        rng=default_rng
    ) -> Trace:
        def sample_site_callback(ps : SampleState):
            assert not ps.fit, f"MetropolisHastings doesn't support Distribution.fit: {ps.name}"
            if ps.name == target_site_name:
                proposal_dist = self.site_proposal_dist(
                    old_value=old_trace[ps.name].value,
                    program_state=ps
                )
                value = proposal_dist.sample(rng=rng)
            elif ps.name in old_trace:
                value = old_trace[ps.name].value
            else:
                value = ps.distribution.sample(rng=rng)
            return value

        new_trace = old_trace.run_from(
            ps=old_trace[target_site_name].program_state,
            old_trace=old_trace,
            sample_site_callback=sample_site_callback,
            observe_site_callback=lambda ps : ps.value,
        )
        return new_trace

    def calc_new_trace_log_probability(
        self,
        old_trace : Trace,
        target_site_name : 'VariableName',
        new_trace : Trace
    ) -> float:
        total_proposal_log_prob = 0
        for entry in new_trace.entries(target_site_name):
            if not entry.is_sample:
                continue
            ps : SampleState = entry.program_state
            if entry.name == target_site_name:
                proposal_dist = self.site_proposal_dist(
                    old_value=old_trace[ps.name].value,
                    program_state=ps
                )
                proposal_log_prob = proposal_dist.log_probability(entry.value)
            elif ps.name in old_trace:
                proposal_log_prob = 0
            else:
                proposal_log_prob = ps.distribution.log_probability(entry.value)
            total_proposal_log_prob += proposal_log_prob
            assert not math.isnan(total_proposal_log_prob)
        return total_proposal_log_prob

    def site_proposal_dist(
        self,
        old_value : 'SampleValue',
        program_state : SampleState,
    ) -> Distribution['SampleValue']:
        if self.custom_proposal_kernels is not None:
            proposal_function = self.custom_proposal_kernels(program_state.name)
            if proposal_function is not None:
                return proposal_function(old_value, program_state)

        if not self.use_drift_kernels:
            return program_state.distribution

        if isinstance(program_state.distribution, Beta):
            if isclose(old_value, 0) or isclose(old_value, 1):
                return program_state.distribution
            # This is the Beta version of the Dirichlet kernel below
            return Beta(
                alpha=old_value*self.simplex_proposal_kernel_alpha,
                beta=(1 - old_value)*self.simplex_proposal_kernel_alpha
            )
        elif isinstance(program_state.distribution.support, Interval):
            # Note that this should be automatically clipped for values out of
            # bounds by setting the score to -inf
            return Uniform(
                old_value - self.uniform_drift_kernel_width/2,
                old_value + self.uniform_drift_kernel_width/2
            )
        elif isinstance(program_state.distribution.support, Simplex):
            # This implementation is borrowed from WebPPL
            return Dirichlet([
                v*self.simplex_proposal_kernel_alpha for v in old_value
            ])
        elif isinstance(program_state.distribution, FiniteDistribution):
            return Categorical(program_state.distribution.support)
        else:
            return program_state.distribution



import math
from flippy import condition, flip
# from flippy.distributions.scipy_dists import Bernoulli, Distribution, Normal,  Gamma, Uniform, Beta
from flippy.distributions.builtin_dists import Categorical, Dirichlet, Bernoulli, Normal, Gamma, Uniform, Beta
from flippy.inference import SamplePrior, SimpleEnumeration, LikelihoodWeighting, MetropolisHastings
from flippy.tools import isclose
from flippy.interpreter import CPSInterpreter, ReturnState, SampleState, ObserveState
from flippy.inference.mcmc.trace import Trace

from flippy.inference.mcmc.metropolis_hastings import MetropolisHastings as MH

def test_mcmc_trace_and_acceptance_ratio():
    def fn():
        if Bernoulli(0.5).sample(name='choice'):
            return Categorical(range(2)).sample(name='rv')
        else:
            Bernoulli(.8).observe(True, name='obs')
            return Categorical(range(3)).sample(name='rv')

    init_ps = CPSInterpreter().initial_program_state(fn).step()

    # test that trace score computation is correct
    tr = Trace.run_from(
        ps=init_ps,
        sample_site_callback=lambda ps : {
            "choice": 0,
            "rv": 0
        }[ps.name],
        observe_site_callback=lambda ps : ps.value,
        old_trace=None
    )
    assert sum(e.is_sample for e in tr.entries()) == 2
    assert sum(e.score for e in tr.entries() if e.is_sample) == math.log(1/2 * 1/3)
    assert sum(e.score for e in tr.entries() if not e.is_sample) == math.log(.8)

    # test that proposal scores are correct
    mh = MH(
        None,
        None,
        use_drift_kernels=False
    )
    new_tr = Trace.run_from(
        ps=init_ps,
        sample_site_callback=lambda ps : {
            "choice": 0,
            "rv": 0
        }[ps.name],
        observe_site_callback=lambda ps : ps.value,
        old_trace=None
    )
    _, target_site_proposal_score = mh.choose_target_site(
        trace=tr,
        target_site_name='rv',
    )
    _, trace_proposal_score = mh.choose_new_trace(
        old_trace=tr,
        new_trace=new_tr,
        target_site_name='rv',
    )
    assert target_site_proposal_score == math.log(1/2), '2 possible target sites'
    assert trace_proposal_score == math.log(1/3), '3 possible values'

    _, trace_proposal_score = mh.choose_new_trace(
        old_trace=tr,
        new_trace=new_tr,
        target_site_name='choice',
    )
    assert trace_proposal_score == math.log(1/2), '2 possible values'

    # test acceptance calculation
    log_acceptance_ratio = mh.calc_log_acceptance_ratio(
        new_score=math.log(.5),
        old_score=math.log(.4),
        new_site_score=math.log(.3),
        old_proposal_score=math.log(.2),
        old_site_score=math.log(.15),
        new_proposal_score=math.log(.1),
    )
    isclose(log_acceptance_ratio, (
        math.log((.5*.3*.2)/(.4*.15*.1))
    ))

def test_mcmc_normal_model():
    hyper_mu, hyper_sigma = -1, 1
    obs = [-.75]*10
    sigma = 1
    def normal_model():
        mu = Normal(hyper_mu, hyper_sigma).sample(name='mu')
        Normal(mu, sigma).observe_all(obs)
        condition(-1.25 < mu < -.5)
        return mu

    seed = 1391299
    mcmc_res = MH(
        function=normal_model,
        samples=4000,
        seed=seed
    ).run()

    lw_res = LikelihoodWeighting(
        function=normal_model,
        samples=4000,
        seed=seed
    ).run()
    assert isclose(mcmc_res.expected_value(), lw_res.expected_value(), atol=.01)

def test_mcmc_gamma_model():
    def gamma_model():
        g = Gamma(3, 2).sample(name='g')
        Uniform(0, g**1.3).observe(0)
        condition(.5 < g < 2)
        return g

    seed = 229932
    mcmc_res = MH(
        function=gamma_model,
        samples=3000,
        burn_in=1000,
        thinning=2,
        seed=seed
    ).run()
    lw_res = LikelihoodWeighting(
        function=gamma_model,
        samples=2000,
        seed=seed
    ).run()
    assert isclose(lw_res.expected_value(), mcmc_res.expected_value(), atol=.01)

def test_mcmc_dirichet_model():
    c1_params = [1, 1, 1]
    c1_data = list('ababababacc')*2
    def model():
        c1 = Dirichlet(c1_params).sample(name='c1')
        dist1 = Categorical(support=list('abc'), probabilities=c1)
        [dist1.observe(d) for d in c1_data]
        return c1

    seed = 13842
    exp_c1 = [n + sum([d == c for d in c1_data]) for c, n in zip('abc', c1_params)]
    exp_c1 = [n / sum(exp_c1) for n in exp_c1]

    mcmc_res = MH(
        function=model,
        samples=2000,
        seed=seed
    ).run()
    for i in [0, 1, 2]:
        est_c1_i = mcmc_res.expected_value(lambda e: e[i])
        assert isclose(est_c1_i, exp_c1[i], atol=.01)

def test_mcmc_beta_model():
    c1_data = [1]*10 + [0]*10
    def model():
        p = Beta(3, 2).sample(name='p')
        bern = Bernoulli(p)
        [bern.observe(d) for d in c1_data]
        return p

    seed = 13842
    exp_p = (3+10) / (3+10 + 2+10)

    mcmc_res = MH(
        function=model,
        samples=1000,
        seed=seed
    ).run()
    est_p = mcmc_res.expected_value()
    assert isclose(est_p, exp_p, atol=.01)


def test_mcmc_categorical_branching_model():
    def model():
        if Bernoulli(0.3).sample(name='choice'):
            return Categorical(range(2)).sample(name='rv')
        else:
            return Categorical(range(3)).sample(name='rv')

    seed = 12949124
    mcmc_res = MH(
        function=model,
        samples=10000,
        seed=seed
    ).run()
    enum_res = SimpleEnumeration(model).run()
    assert isclose(mcmc_res.expected_value(), enum_res.expected_value(), atol=.02)

def test_mcmc_geometric():
    def geometric(p):
        '''
        The probability distribution of the number X of Bernoulli trials needed to get one success.
        https://en.wikipedia.org/wiki/Geometric_distribution
        '''
        x = Bernoulli(p).sample()
        if x == 1:
            return 1
        return 1 + geometric(p)

    param = 0.98
    expected = 1/param

    seed = 13852

    mh_res = MH(
        function=geometric,
        samples=10000,
        seed=seed,
    ).run(param)
    assert isclose(mh_res.expected_value(), expected, atol=.01)

def text_mcmc_categorical_branching_explicit_names():
    def fn():
        if Bernoulli(.5).sample(name="choice"):
            x = Categorical(['a', 'b'], probabilities=[.5, .5]).sample(name='x')
        else:
            x = Categorical(['c', 'b'], probabilities=[.8, .2]).sample(name='x')
        return x
    enum_dist = SimpleEnumeration(fn).run()
    mh_dist = MH(fn, samples=10000, seed=124).run()
    for e in enum_dist:
        assert isclose(enum_dist[e], mh_dist[e], atol=1e-2)

def test_mcmc_initial_value():
    def model():
        p = Beta(3, 2).sample(name='p', initial_value=0.123)
        return p

    mcmc = MetropolisHastings(function=model, samples=10)
    init_ps = mcmc.initial_program_state.step()
    trace = mcmc.generate_initial_trace(init_ps)
    assert trace['p'].value == 0.123

def test_mcmc_potential_index_error():
    def fn():
        xs = [0, 1, 2] if flip() else [9, 10]
        idx = Categorical(range(len(xs))).sample()
        return xs[idx]
    enum_dist = SimpleEnumeration(fn).run()
    mh_dist = MH(fn, samples=10000, seed=124).run()
    assert enum_dist.isclose(mh_dist, atol=5e-2)



import linecache
import math
from collections import defaultdict
from typing import Generic

from flippy.core import ReturnState, SampleState, ObserveState
from flippy.types import Element
from flippy.interpreter import CPSInterpreter
from flippy.distributions import Categorical, RandomNumberGenerator
from flippy.inference.inference import InferenceAlgorithm


class LikelihoodWeighting(InferenceAlgorithm[Element]):
    """
    Likelihood weighting inference algorithm. Samples from the prior
    and weights samples by the likelihood of observed data.
    - `function`: The function to be executed
    - `samples`: The number of samples to draw
    - `seed`: Optional random seed for reproducibility
    - `_cpus`: Number of CPUs to use for parallel execution (default is 1)
    """
    def __init__(
        self,
        function,
        samples : int,
        seed=None,
        _cpus=1,
        _joblib_backend='loky'
    ):
        self.function = function
        self.samples = samples
        self.seed = seed
        self._cpus = _cpus
        self._joblib_backend = _joblib_backend

    @property
    def is_cachable(self):
        return self.seed is not None

    def run(self, *args, **kws) -> Categorical[Element]:
        if self._cpus == 1:
            return_counts = self._run_batch(
                *args, **kws,
                samples=self.samples,
                seed=self.seed,
                _linecache=linecache.cache
            )
        else:
            return_counts = self._run_parallel(*args, **kws)
        total_prob = sum(return_counts.values())
        return_probs = {e: p/total_prob for e, p in return_counts.items()}
        return Categorical.from_dict(return_probs)

    def _run_batch(self, *args, samples: int, seed: int, _linecache, **kws):
        # restore linecache so inspect.getsource works for interactively defined functions
        linecache.cache = _linecache

        rng = RandomNumberGenerator(seed)
        init_ps = CPSInterpreter().initial_program_state(self.function)
        return_counts = defaultdict(float)
        for _ in range(samples):
            weight = 0
            ps = init_ps.step(*args, **kws)
            while not isinstance(ps, ReturnState):
                if isinstance(ps, SampleState):
                    assert not ps.fit, f"LikelihoodWeighting doesn't support Distribution.fit: {ps.name}"
                    value = ps.distribution.sample(rng=rng)
                    ps = ps.step(value)
                elif isinstance(ps, ObserveState):
                    weight += ps.distribution.log_probability(ps.value)
                    if weight == float('-inf'):
                        break
                    ps = ps.step()
                else:
                    raise ValueError("Unrecognized program state")
            if weight == float('-inf'):
                continue
            return_counts[ps.value] += math.exp(weight)
        return return_counts

    def _run_parallel(self, *args, **kws):
        from joblib import Parallel, delayed, cpu_count
        rng = RandomNumberGenerator(self.seed)
        if self._cpus == -1:
            cpus = cpu_count()
        else:
            cpus = self._cpus
        all_return_counts = Parallel(n_jobs=cpus, backend=self._joblib_backend)(
            delayed(self._run_batch)(
                *args, **kws,
                samples=self.samples // cpus,
                seed=rng.new_seed(),
                _linecache=linecache.cache
            )
        for _ in range(self._cpus))
        return_counts = self._run_batch(
            *args, **kws,
            samples=self.samples % cpus,
            seed=rng.new_seed(),
            _linecache=linecache.cache,
        )
        for rc in all_return_counts:
            for k, v in rc.items():
                return_counts[k] += v
        return return_counts



from typing import Tuple
import linecache

from flippy.core import ReturnState, SampleState, ObserveState, ProgramState
from flippy.distributions.random import default_rng, RandomNumberGenerator
from flippy.distributions import ZeroDistributionError
from flippy.interpreter import CPSInterpreter
from flippy.distributions import Categorical, RandomNumberGenerator
from flippy.distributions.support import Interval
from flippy.inference import Enumeration
from flippy.types import Element
from flippy.inference.inference import InferenceAlgorithm, MarginalLikelihood, \
    DiscreteInferenceResult
from flippy.tools import PackagePlaceholder

try:
    import scipy.optimize as scipy_optimize
except ImportError:
    scipy_optimize = PackagePlaceholder("scipy.optimize")

try:
    import numpy as np
except ImportError:
    np = PackagePlaceholder("numpy")

try:
    import joblib
except ImportError:
    joblib = PackagePlaceholder("joblib")


class MaximumMarginalAPosteriori(InferenceAlgorithm[Element]):
    """
    Maximum Marginal A Posteriori (MMAP) inference algorithm. Finds the assignment
    to continuous variables that maximizes the marginal posterior probability of
    discrete variables.

    - `function`: The function to be executed
    - `seed`: Optional random seed for reproducibility
    - `method`: Scipy optimization method to use (see
    [scipy docs](https://docs.scipy.org/doc/scipy/reference/optimize.html)
    for options)
    - `disp`: Whether to display optimization output
    - `maxiter`: Maximum number of iterations for the optimizer
    - `maxfev`: Maximum number of function evaluations for the optimizer
    - `maximum_likelihood`: If True, finds the assignment that maximizes the
    likelihood of the data instead of the posterior
    - `min_valid_runs`: Minimum number of valid optimization runs to consider
    - `max_retries`: Maximum number of optimization runs to attempt
    - `_cpus`: Number of CPUs to use for parallel execution (default is 1)
    - `**enumeration_kwargs`: Additional keyword arguments to pass to the
    Enumeration inference algorithm
    """
    def __init__(
        self,
        function,
        seed=None,
        method="Powell",
        disp=False,
        maxiter=1000,
        maxfev=1000,
        maximum_likelihood=False,
        min_valid_runs=1,
        max_retries=10,
        _cpus=1,
        **enumeration_kwargs
    ):
        self.function = function
        self.enumeration_alg = Enumeration(
            function,
            **enumeration_kwargs
        )
        self.seed = seed
        self.method = method
        self.maxiter = maxiter
        self.maxfev = maxfev
        self.disp = disp
        self.maximum_likelihood = maximum_likelihood
        self.min_valid_runs = min_valid_runs
        self.max_retries = max_retries
        self._cpus = _cpus

    @property
    def is_cachable(self):
        return self.seed is not None

    def single_run(self, *args, **kwargs) -> Tuple[MarginalLikelihood, DiscreteInferenceResult[Element]]:
        rng = RandomNumberGenerator(self.seed)
        done = False
        assignment = {}
        while not done:
            assignment, done = self.optimize_with_assignment(args, kwargs, assignment, rng=rng)
        _, score, result = self.assignment_score(args, kwargs, assignment, rng=rng)
        if result is None:
            raise ZeroDistributionError
        return score, result

    def _run(self, *args, **kwargs) -> Tuple[float, DiscreteInferenceResult[Element]]:
        def job(linecache_cache) -> Tuple[float, DiscreteInferenceResult[Element]]:
            linecache.cache.update(linecache_cache)
            try:
                return self.single_run(*args, **kwargs)
            except ZeroDistributionError:
                return float('-inf'), None
        assert self._cpus > 0
        with joblib.Parallel(n_jobs=self._cpus) as parallel:
            valid_results = []
            retries = 0
            while len(valid_results) < self.min_valid_runs:
                results = parallel(
                    joblib.delayed(job)(linecache.cache) for _ in range(self._cpus)
                )
                retries += len(results)
                valid_results.extend([r for r in results if r[0] > float('-inf')])
                if retries >= self.max_retries:
                    break
        if len(valid_results) == 0:
            raise ZeroDistributionError("Unable to find minimum number of valid solutions.")
        best_score, best_result = max(valid_results, key=lambda x: x[0])
        metadata = dict(retries=retries, n_valid_results=len(valid_results))
        return best_score, best_result, metadata

    def run(self, *args, **kwargs) -> 'MaximumMarginalAPosterioriResult[Element]':
        best_score, best_result, metadata = self._run(*args, **kwargs)
        values, probs = zip(*best_result.items())
        result = MaximumMarginalAPosterioriResult(values, probs, np.exp(best_score))
        result.update_metadata(metadata)
        return result

    def optimize_with_assignment(self, args, kwargs, assignment: dict, rng=default_rng):
        init_assignments, _, _ = self.assignment_score(args, kwargs, assignment, rng=rng)
        varnames = list(init_assignments.keys())
        init_values = [init_assignments[v][0] for v in varnames]
        bounds = [init_assignments[v][1] for v in varnames]
        def objective(x):
            assignment = dict(zip(varnames, zip(x, bounds)))
            new_assignment, log_marginal_likelihood, _ = \
                self.assignment_score(args, kwargs, assignment, rng=rng)
            if len(new_assignment) > len(assignment):
                raise NewVariableException(new_assignment)
            return -np.exp(log_marginal_likelihood)
        try:
            res = scipy_optimize.minimize(
                objective,
                init_values,
                method=self.method,
                options={
                    "disp": self.disp,
                    "maxiter": self.maxiter,
                    "maxfev": self.maxfev
                },
                bounds=bounds
            )
            new_assignment = dict(zip(varnames, zip(res.x, bounds)))
            done = res.success
        except NewVariableException as res:
            new_assignment = res.args[0]
            done = False
        return new_assignment, done

    def assignment_score(self, args, kwargs, assignments: dict, rng=default_rng):
        assert all(isinstance(v, tuple) for v in assignments.values())
        assignments = assignments.copy()
        cont_var_sample_score = [0]
        def cont_var_func(ps: SampleState):
            if ps.name not in assignments:
                assert isinstance(ps.distribution.support, Interval)
                if ps.initial_value in ps.distribution.support:
                    value = ps.initial_value
                else:
                    value = ps.distribution.sample(rng=rng)
                bounds = ps.distribution.support.start, ps.distribution.support.end
                assignments[ps.name] = (value, bounds)
            value, bounds = assignments[ps.name]
            cont_var_sample_score[0] += ps.distribution.log_probability(value)
            return [value]
        self.enumeration_alg.set_cont_var_func(cont_var_func)
        try:
            result = self.enumeration_alg.run(*args, **kwargs)
            score = np.log(result.marginal_likelihood) if result.marginal_likelihood > 0 else float("-inf")
        except ZeroDistributionError:
            result = None
            score = float("-inf")
        if self.maximum_likelihood:
            score -= cont_var_sample_score[0]
        return assignments, score, result

class NewVariableException(Exception):
    pass

class MaximumMarginalAPosterioriResult(DiscreteInferenceResult[Element]):
    retries : int
    n_valid_results : int

    def update_metadata(self, metadata: dict):
        self.retries = metadata.get("retries", None)
        self.n_valid_results = metadata.get("n_valid_results", None)



import math

import pytest
import numpy as np

from flippy.distributions import Bernoulli, Categorical, Dirichlet, Uniform, \
    Normal, Gamma, Beta, MultivariateNormal, InverseWishart, NormalNormal, \
    MultivariateNormalNormal
from flippy.distributions.support import Interval, Range, ProductSet, \
    UnionSet, Simplex, OrderedIntegerPartitions
from flippy.inference.likelihood_weighting import LikelihoodWeighting
from flippy.inference.enumeration import Enumeration
from flippy.interpreter import CPSInterpreter
from flippy.core import ReturnState
from flippy.tools import isclose
from flippy.distributions.random import default_rng, RandomNumberGenerator

def test_scipy_uniform():
    dist = Uniform(-1, -.5)
    for i in range(100):
        u = dist.sample()
        assert -1 <= u <= -.5
        assert isclose(dist.log_probability(u), math.log(1/.5))

def test_distribution_bool():
    dist = Uniform(-1, -.5)
    with pytest.raises(ValueError):
        bool(dist)

def test_normal_normal():
    hyper_mu, hyper_sigma = -1, 4
    obs = [-.75]*10
    sigma = 1
    def normal_model():
        mu = Normal(hyper_mu, hyper_sigma).sample(name='mu')
        dist = Normal(mu, sigma)
        dist.observe_all(obs)
        return mu

    seed = 2391299
    lw_res = LikelihoodWeighting(
        function=normal_model,
        samples=4000,
        seed=seed
    ).run()
    nn = NormalNormal(prior_mean=hyper_mu, prior_sd=hyper_sigma, sd=sigma)

    assert isclose(lw_res.expected_value(), nn.update(obs).prior_mean, atol=.01)


def test_multivariate_normal_multivariate_normal():
    mean = default_rng.random()
    priorvar = default_rng.random()
    sigma2 = default_rng.random()

    mvn = MultivariateNormalNormal(
        prior_means=[mean,mean],
        prior_cov=[[priorvar,0],[0,priorvar]],
        cov=[[sigma2,0],[0,sigma2]]
    )

    sample = mvn.sample()
    uvn = NormalNormal(prior_mean=mean, prior_sd=priorvar**.5, sd=sigma2**.5)
    uvnlogprob = uvn.total_log_probability(sample.flatten())
    mvnlogprob = mvn.log_probability(sample)

    assert isclose(uvnlogprob, mvnlogprob)


def test_multivariate_normal():
    rng = RandomNumberGenerator(12345)
    mvn = MultivariateNormal(means=(0, 1), covariance=((1, 0.5), (0.5, 1)))
    samples = np.array([mvn.sample(rng=rng) for _ in range(5000)])
    sample_means = np.mean(samples, axis=0)
    sample_cov = np.prod(samples - sample_means, axis=1).mean()
    assert (np.abs(sample_means - (0, 1)) < .02).all()
    assert abs(sample_cov - 0.5) < 1e-2


def test_InverseWishart():
    invwish = InverseWishart(df=4.2, scale_matrix=((1, 0.5), (0.5, 1)))
    rng = RandomNumberGenerator(32345)
    samples = [invwish.sample(rng=rng) for _ in range(5000)]
    assert np.isclose(
        np.array(samples).mean(axis=0),
        invwish.expected_value(),
        atol=0.1
    ).all()

def test_categorical_dist_equality():
    def f():
        if Bernoulli().sample():
            return Categorical(['A', 'B'])
        return Categorical(['A', 'B'])
    dist = Enumeration(f).run()
    assert dist.isclose(Categorical([
        Categorical(['A', 'B'], probabilities=[.5, .5])
    ]))

def test_categorical_condition():
    d = Categorical(range(10))
    d = d.condition(lambda x : x % 2 == 0)
    d = d.condition(lambda x : x % 3 == 0)
    assert d.isclose(Categorical([0, 6], probabilities=[.5, .5]))

def test_observe_all():
    def model_observe_all(p, data):
        Bernoulli(p).observe_all(data)
        return p

    def model_observe(p, data):
        [Bernoulli(p).observe(d) for d in data]
        return p

    data = (1, 1, 1, 1, 1, 0, 0)
    p = .9

    ps = CPSInterpreter().initial_program_state(model_observe_all)
    ps = ps.step(p, data)
    logprob1 = ps.distribution.log_probability(ps.value)

    ps = CPSInterpreter().initial_program_state(model_observe)
    ps = ps.step(p, data)
    logprob2 = 0
    while not isinstance(ps, ReturnState):
        logprob2 += ps.distribution.log_probability(ps.value)
        ps = ps.step()
    assert isclose(logprob1, logprob2)

def test_Dirichlet_numerical_stability():
    rng = RandomNumberGenerator(12345)
    for _ in range(100):
        dist = Dirichlet([.1, .0001, 4.0])
        p = dist.sample(rng=rng)
        assert 0 not in p, "We should not be able to sample a vector with a 0"
        assert 1 not in p, "We should not be able to sample a vector with a 1"
        assert dist.log_probability(p) > float('-inf')

def test_support():
    closed_interval = Interval(0, 1, left_open=False, right_open=False)
    assert 0.0 in closed_interval
    assert 1.0 in closed_interval

    open_interval = Interval(0, 1, left_open=True, right_open=True)
    assert .00001 in open_interval
    assert .99999 in open_interval
    assert 0.0 not in open_interval
    assert 1.0 not in open_interval

    half_open_interval = Interval(0, 1, left_open=True, right_open=False)
    assert .00001 in half_open_interval
    assert .99999 in half_open_interval
    assert 0.0 not in half_open_interval
    assert 1.0 in half_open_interval

    range1 = Range(10)
    assert 0 in range1
    assert 9 in range1
    assert 10 not in range1

    range2 = Range(5, 10)
    assert 1 not in range2
    assert 5 in range2
    assert 9 in range2
    assert 10 not in range2

    range3 = Range(5, 10, 2)
    assert 5 in range3
    assert 7 in range3
    assert 9 in range3
    assert 6 not in range3
    assert 10 not in range3
    assert 11 not in range3

    product_set_1 = ProductSet(Range(0, 2), Range(5, 7))
    assert (0, 5) in product_set_1
    assert (1, 6) in product_set_1
    assert (2, 5) not in product_set_1
    assert (0, 7) not in product_set_1
    assert (1, 5) in product_set_1
    assert 1 not in product_set_1

    product_set_2 = ProductSet(Range(10), ("A", "B", "C"))
    assert (0, "A") in product_set_2
    assert (9, "C") in product_set_2
    assert (10, "A") not in product_set_2

    product_set_3 = ProductSet(Range(0, 2), Interval(0, 2))
    assert (0, 0.5) in product_set_3
    assert (1, 1.5) in product_set_3
    assert (2, 0.5) not in product_set_3

    union_set_1 = UnionSet(Range(0, 2), Range(5, 7))
    assert 0 in union_set_1
    assert 1 in union_set_1
    assert 5 in union_set_1
    assert 6 in union_set_1
    assert 2 not in union_set_1
    assert 4 not in union_set_1

    simplex_3 = Simplex(3)
    assert (0.5, 0.5, 0.0) in simplex_3
    assert (0.3, 0.4, 0.3) in simplex_3
    assert (0.1, 0.2, 0.6) not in simplex_3
    assert (.5, 0.5, 0.1) not in simplex_3

    ordered_partitions = OrderedIntegerPartitions(5, 3)
    assert (5, 0, 0) in ordered_partitions
    assert (4, 1, 0) in ordered_partitions
    assert (3, 2, 0) in ordered_partitions
    assert (2, 1, 1) not in ordered_partitions



import heapq
import linecache
from dataclasses import dataclass
from collections import defaultdict
from typing import Any, Union, List, Tuple, Dict, Set, Callable, Generic
from functools import cached_property

from flippy.core import ProgramState, ReturnState, SampleState, ObserveState, InitialState, GlobalStore
from flippy.interpreter import CPSInterpreter
from flippy.types import Element
from flippy.transforms import CPSTransform
from flippy.distributions import Categorical, ZeroDistributionError
from flippy.tools import logsumexp
from flippy.callentryexit import EnterCallState, ExitCallState
from flippy.map import MapEnter, MapExit
from flippy.inference.inference import InferenceAlgorithm, DiscreteInferenceResult
from flippy.hashable import hashabledict
from flippy.tools import LRUCache, PackagePlaceholder
from flippy.distributions.support import Interval

try:
    import joblib
except ImportError:
    joblib = PackagePlaceholder("joblib")

@dataclass
class ScoredProgramState:
    cumulative_score: float
    program_state: tuple

    def __lt__(self, other: 'ScoredProgramState'):
        return self.cumulative_score > other.cumulative_score

    def __eq__(self, other: 'ScoredProgramState'):
        return self.cumulative_score == other.cumulative_score

    def __gt__(self, other: 'ScoredProgramState'):
        return self.cumulative_score < other.cumulative_score

    def __iter__(self):
        return iter((self.cumulative_score, self.program_state))

@dataclass
class ExecutionResult:
    states: List[ProgramState]
    scores: List[float]
    cont_var_assigned: bool = False

    def collapse(self):
        if len(self) == 0:
            return self
        assert len(self.states) == len(self.scores)
        states_scores = defaultdict(lambda : float('-inf'))
        for state, score in zip(self.states, self.scores):
            states_scores[state] = logsumexp(states_scores[state], score)
        states, scores = zip(*states_scores.items())
        return ExecutionResult(states, scores, **self.info)

    def add_score(self, score):
        new_scores = [s + score for s in self.scores]
        return ExecutionResult(self.states, new_scores, **self.info)

    @property
    def info(self) -> Dict[str, Any]:
        return dict(cont_var_assigned=self.cont_var_assigned)

    def __iter__(self):
        yield from zip(self.states, self.scores)

    def __len__(self):
        assert len(self.states) == len(self.scores)
        return len(self.states)

@dataclass
class ReturnExecutionBranch(ExecutionResult):
    states: List[Union[ReturnState, ExitCallState]]
    scores: List[float]

class Enumeration(InferenceAlgorithm,Generic[Element]):
    """
    Enumerates all possible executions of a program with discrete random variables
    as a graph by identifying call entry and exit points and reusing
    cached results when possible.

    - `function`: The function to be executed
    - `max_states`: The maximum number of states to consider
    - `cont_var_func`: A function to provide continuous variable values for sample states
    - `_call_cache_size`: The size of the call cache
    - `_cpus`: The number of CPUs to use for parallel execution
    - `_emit_call_entryexit`: Whether to emit call entry/exit states
    - `_state_visit_callback`: A callback function to track state visits
    - `_track_sample_state_visits`: Whether to track visits to sample states
    """

    is_cachable = True  # Inference results can always be cached

    def __init__(
        self,
        function,
        max_states=float('inf'),
        _call_cache_size=128,
        _cpus=1,
        _emit_call_entryexit=True,
        _state_visit_callback: Callable[[ProgramState], None]=None,
        _track_sample_state_visits=False,
        cont_var_func : Callable[[SampleState], List] = None
    ):
        assert _cpus > 0, "Number of CPUs must be greater than 0"
        self.cps = CPSInterpreter(_emit_call_entryexit=_emit_call_entryexit)
        if not CPSTransform.is_transformed(function):
            function = self.cps.non_cps_callable_to_cps_callable(function)
        self.function = function
        self.max_states = max_states
        self._max_states_per_partition = max_states // _cpus if max_states < float('inf') else float('inf')
        self._track_sample_state_visits = _track_sample_state_visits or (max_states < float('inf'))
        if _call_cache_size > 0:
            self._call_cache = LRUCache(max_size=_call_cache_size)
        else:
            self._call_cache = None
        self._cpus = _cpus
        self._emit_call_entryexit = _emit_call_entryexit
        self._cont_var_func = cont_var_func
        self._n_sample_states_visited = None
        self._state_visit_callback = self._create_state_visit_callback(_state_visit_callback)


    def _create_state_visit_callback(
        self,
        _state_visit_callback: Callable[[ProgramState], None]=None
    ):
        if self._track_sample_state_visits:
            def increment_state_visit_count(ps):
                if isinstance(ps, SampleState):
                    self._n_sample_states_visited += 1
                if _state_visit_callback is not None:
                    _state_visit_callback(ps)
            return increment_state_visit_count


    @cached_property
    def init_ps(self):
        return self.cps.initial_program_state(self.function)


    def set_cont_var_func(self, f: Callable[[SampleState], List]):
        self._cont_var_func = f


    def run(self, *args, **kws) -> DiscreteInferenceResult[Element]:
        return_values, return_scores = self._run(*args, **kws)
        return DiscreteInferenceResult.from_values_scores(return_values, return_scores)


    def _run(self, *args, **kws):
        if self._track_sample_state_visits:
            self._n_sample_states_visited = 0
        if self._cpus == 1:
            return_values, return_scores = self._run_single(*args, **kws)
        else:
            return_values, return_scores = self._run_parallel(*args, **kws)
        return return_values, return_scores


    def _run_single(self, *args, **kws):
        ps, score = self.next_choice_state(self.init_ps, args=args, kwargs=kws)
        result = self.enumerate_return_states_scores(ps)
        if len(result) == 0:
            raise ZeroDistributionError
        result = result.add_score(score)
        return_vals = [rs.value for rs in result.states]
        return return_vals, result.scores


    def _run_parallel(self, *args, **kws):
        assert CPSTransform.is_transformed(self.function), \
            "Function must be CPS transformed prior to creating workers"
        if self._cpus < 0:
            cpus = joblib.cpu_count() + 1 + self._cpus
        else:
            cpus = self._cpus

        all_value_scores = joblib.Parallel(n_jobs=cpus, backend="loky")(
            joblib.delayed(self._run_partition)(
                *args, **kws,
                _partition_idx=i,
                _partitions=cpus,
                _linecache=linecache.cache,
            )
        for i in range(cpus))
        values = sum([s for s, _ in all_value_scores], [])
        scores = sum([s for _, s in all_value_scores], [])
        return values, scores


    def _run_partition(self, *args, _partition_idx, _partitions, _linecache, **kws):
        linecache.cache = _linecache # restore linecache so inspect.getsource works
        ps, init_score = self.next_choice_state(self.init_ps, args=args, kwargs=kws)
        result = self.enumerate_return_states_scores(
            init_ps=ps,
            _partition_idx=_partition_idx,
            _partitions=_partitions
        )
        result = result.add_score(init_score)
        values = [state.value for state in result.states]
        return values, result.scores


    def enumerate_return_states_scores(
        self,
        init_ps: ProgramState,
        _partition_idx=0,
        _partitions=1
    ) -> ReturnExecutionBranch:
        assert _partition_idx < _partitions, "Partition index must be less than the number of partitions"
        if isinstance(init_ps, (ReturnState, ExitCallState)):
            return ReturnExecutionBranch([init_ps], [0.])
        frontier: List[ScoredProgramState] = []
        return_states = []
        return_scores = []
        heapq.heappush(frontier, ScoredProgramState(0., init_ps))
        in_partition = _partitions == 1
        result = None
        while len(frontier) > 0:
            if not in_partition and len(frontier) >= _partitions:
                # This block of code only occurs once in each worker during
                # multi-cpu enumeration (and never in single-cpu enumeration).
                # We partition trace space by expanding the frontier until there
                # are at least as many program states as there are partitions.
                # Then we assign a subset of the frontier to each worker.
                if _partition_idx != 0:
                    return_states, return_scores = [], []
                sub_frontier = []
                for i, ps in enumerate(frontier):
                    if i % _partitions == _partition_idx:
                        sub_frontier.append(ps)
                frontier = sub_frontier
                in_partition = True
                continue

            if (
                self._track_sample_state_visits and \
                self._n_sample_states_visited >= self._max_states_per_partition
            ):
                return ReturnExecutionBranch(return_states, return_scores)

            cum_score, ps = heapq.heappop(frontier)
            result = self.enumerate_successors_scores(ps)

            for new_ps, score in result:
                if score == float('-inf'):
                    continue
                if isinstance(new_ps, (ReturnState, ExitCallState)):
                    return_states.append(new_ps)
                    return_scores.append(cum_score + score)
                else:
                    heapq.heappush(frontier, ScoredProgramState(cum_score + score, new_ps))

        return ReturnExecutionBranch(return_states, return_scores, **result.info)


    def enumerate_successors_scores(
        self,
        ps: ProgramState,
    ) -> ExecutionResult:
        # we enumerate successors of a program state differently depending on
        # what kind of state it is
        if isinstance(ps, SampleState):
            return self.enumerate_sample_state_successors(ps)
        elif isinstance(ps, EnterCallState):
            result = self.enumerate_enter_call_state_successors(ps)
            # we need to take the next step for each successor
            successors = []
            scores = []
            for exit_state, exit_score in result:
                new_ps, new_score = self.next_choice_state(exit_state)
                if new_score == float('-inf'):
                    continue
                successors.append(new_ps)
                scores.append(exit_score + new_score)
            return ExecutionResult(successors, scores, **result.info)
        elif isinstance(ps, MapEnter):
            raise NotImplementedError
        elif isinstance(ps, InitialState):
            new_ps, step_score = self.next_choice_state(ps)
            if step_score > float('-inf'):
                successors = [new_ps]
                scores = [step_score]
            else:
                successors = []
                scores = []
            return ExecutionResult(successors, scores)
        else:
            raise ValueError(f"Unrecognized program state message {ps}")


    def enumerate_sample_state_successors(
        self,
        ps: SampleState,
    ) -> ExecutionResult:
        if isinstance(ps.distribution.support, Interval):
            assert self._cont_var_func is not None, \
                "Continuous sample state values must be provided for continuous distributions"
            values = self._cont_var_func(ps)
            cont_var_assigned = True
        else:
            assert not ps.fit, f"Enumeration doesn't support Distribution.fit: {ps.name}"
            values = ps.distribution.support
            cont_var_assigned = False
        successors, scores = [], []
        for value in values:
            score = ps.distribution.log_probability(value)
            new_ps, new_score = self.next_choice_state(ps, value=value)
            if new_score == float('-inf'):
                continue
            score += new_score
            successors.append(new_ps)
            scores.append(score)
        return ExecutionResult(successors, scores, cont_var_assigned=cont_var_assigned)

    def enumerate_enter_call_state_successors(
        self,
        enter_state: EnterCallState,
    ) -> ExecutionResult:
        if self._call_cache is None:
            return self._enumerate_enter_call_state_successors(enter_state)

        # we never cache the root call
        is_root_call = len(enter_state.stack) == 1
        if is_root_call:
            return self._enumerate_enter_call_state_successors(enter_state)

        # This logic handles caching using an LRU cache
        global_store_key = hashabledict(enter_state.init_global_store.store)
        key = (enter_state.function, enter_state.args, enter_state.kwargs, global_store_key)
        if key in self._call_cache:
            exit_values, exit_scores = self._call_cache[key]
            exit_states = []
            for rv, gs in exit_values:
                gs: GlobalStore
                next_state = enter_state.skip(rv)
                next_state.set_init_global_store(gs.copy(), force=True)
                exit_states.append(next_state)
            result = ExecutionResult(exit_states, exit_scores)
            if self._track_sample_state_visits:
                self._n_sample_states_visited += len(exit_states) > 1
        else:
            result = self._enumerate_enter_call_state_successors(enter_state)
            # we can only cache if no continuous variables were assigned during the call
            if not result.cont_var_assigned:
                exit_values = [(rs.value, rs.init_global_store) for rs in result.states]
                self._call_cache[key] = (exit_values, result.scores)
        return result

    def _enumerate_enter_call_state_successors(
        self,
        init_ps: EnterCallState
    ) -> ExecutionResult:
        # when we enter a call, we take the first step, see if it halts or returns immediately
        assert isinstance(init_ps, EnterCallState)
        ps, init_score = self.next_choice_state(init_ps)
        if init_score == float('-inf'):
            return ExecutionResult([], [])
        if isinstance(ps, ExitCallState):
            return ExecutionResult([ps], [init_score])

        # if it doesn't immediately exit, we enumerate to get all the exit states/scores
        result = self.enumerate_return_states_scores(init_ps=ps)
        result = result.collapse()
        result = result.add_score(init_score)
        return result


    def next_choice_state(
        self,
        init_ps: ProgramState,
        *,
        value=None,
        args=None,
        kwargs=None,
    ) -> Tuple[ProgramState, float]:
        """
        This runs a program state until it reaches a choice state or an exit
        state. It returns the next choice state and score collected along the way.
        """

        if isinstance(init_ps, ObserveState):
            score = init_ps.distribution.log_probability(init_ps.value)
        else:
            score = 0.
        if isinstance(init_ps, (SampleState, MapExit)):
            ps = init_ps.step(value)
        elif isinstance(init_ps, InitialState):
            ps = init_ps.step(*args, **kwargs)
            # We skip the call entry state associated with the root call
            # since (1) we don't want to cache it and (2) don't want to consider it
            # for partitioning across workers
            if isinstance(ps, EnterCallState) and ps.is_root_call:
                ps = ps.step()
        else:
            ps = init_ps.step()
        while not isinstance(ps, (SampleState, EnterCallState, ReturnState, ExitCallState)):
            if self._state_visit_callback:
                self._state_visit_callback(ps)
            if isinstance(ps, ObserveState):
                score += ps.distribution.log_probability(ps.value)
            if score == float('-inf'):
                return None, float('-inf')
            ps = ps.step()
        if self._state_visit_callback:
            self._state_visit_callback(ps)
        return ps, score




import random

MAX_INT64 = 2**64 - 1

class BaseRandomNumberGenerator(random.Random):
    def new_seed(self):
        return self.randint(0, MAX_INT64)

try:
    import numpy as np
    class RandomNumberGenerator(BaseRandomNumberGenerator):
        """
        Random number generator for both Python and NumPy.

        To access the NumPy random generator, use `rng.np`.
        """
        def __init__(self, seed=None):
            super().__init__(seed)
            self.np = np.random.Generator(np.random.PCG64(self.randint(0, MAX_INT64)))
except ImportError:
    RandomNumberGenerator = BaseRandomNumberGenerator

default_rng = RandomNumberGenerator(None)



from flippy.tools import isclose, LRUCache

def test_isclose():
    assert not isclose(1, 1+1e-1)
    assert not isclose(1, 1+1e-2)
    assert not isclose(1, 1+1e-3)
    assert not isclose(1, 1+1e-4)
    assert isclose(1, 1+1e-5)
    assert isclose(1, 1+1e-6)
    assert isclose(1, 1+1e-7)
    assert isclose(1, 1+1e-8)

def test_lru_cache():
    lru_inf = LRUCache(max_size=None)
    assert lru_inf.max_size == float('inf')

    lru_finite = LRUCache(max_size=2)
    lru_finite['a'] = 1
    lru_finite['b'] = 2
    lru_finite['c'] = 3
    assert set(lru_finite.keys()) == {'b', 'c'}
    lru_finite['d'] = 4
    assert set(lru_finite.keys()) == {'c', 'd'}
    lru_finite['c']
    lru_finite['e'] = 5
    assert set(lru_finite.keys()) == {'c', 'e'}
    lru_finite['e'] = 6
    assert lru_finite['e'] == 6



import collections
import math
import pytest

from flippy import flip, mem, condition, draw_from
from flippy.distributions.builtin_dists import \
    Bernoulli, Distribution, Categorical, Gaussian, Uniform, Binomial, Poisson
from flippy.inference import SamplePrior, SimpleEnumeration, LikelihoodWeighting, _distribution_from_inference
from flippy.inference.inference import DiscreteInferenceResult
from flippy.inference.enumeration import Enumeration
from flippy.inference.max_marg_post import MaximumMarginalAPosteriori
from flippy.tools import isclose
from flippy.interpreter import CPSInterpreter, ReturnState, SampleState, ObserveState
from flippy.callentryexit import register_call_entryexit
from flippy.map import independent_map

def geometric(p):
    '''
    The probability distribution of the number X of Bernoulli trials needed to get one success.
    https://en.wikipedia.org/wiki/Geometric_distribution
    '''
    x = Bernoulli(p).sample()
    if x == 1:
        return 1
    return 1 + geometric(p)

def geometric_iter(p):
    ct = 1
    while Bernoulli(p).sample() == 0:
        ct += 1
    return ct

def expectation(d: Distribution, projection=lambda s: s):
    total = 0
    partition = 0
    for s in d.support:
        p = math.exp(d.log_probability(s))
        total += p * projection(s)
        partition += p
    assert isclose(partition, 1)
    return total

def test_enumeration_geometric():
    param = 0.25
    expected = 1/param
    for fn in [geometric, geometric_iter]:
        rv = SimpleEnumeration(fn, max_executions=100).run(param)
        d = _distribution_from_inference(rv)
        assert isclose(expectation(d), expected)

        assert len(rv) == 100
        assert set(rv.keys()) == set(range(1, 101)), set(rv.keys()) - set(range(1, 101))
        for k, sampled_prob in rv.items():
            pmf = (1-param) ** (k - 1) * param
            # This will only be true when executions is high enough, since
            # sampled_prob is normalized.
            assert isclose(sampled_prob, pmf), (k, sampled_prob, pmf)

def test_likelihood_weighting_and_sample_prior():
    param = 0.98
    expected = 1/param

    seed = 13842

    lw_dist = LikelihoodWeighting(geometric, samples=1000, seed=seed).run(param)
    lw_exp = expectation(_distribution_from_inference(lw_dist))
    prior_dist = SamplePrior(geometric, samples=1000, seed=seed).run(param)
    prior_exp = expectation(_distribution_from_inference(prior_dist))

    assert lw_exp == prior_exp, 'Should be identical when there are no observe statements'

    assert isclose(expected, prior_exp, atol=1e-2), 'Should be somewhat close to expected value'

import numpy as np
def cosine_similarity(a, b):
    a = np.array(a)
    b = np.array(b)
    return (a@b)/((a**2).sum() * (b**2).sum())**.5

def test_observations():
    def model_simple():
        rv = Categorical(range(3)).sample()
        Bernoulli(2**(-rv)).observe(True)
        return rv

    def model_branching():
        if Bernoulli(0.5).sample(name='choice'):
            Bernoulli(.2).observe(True, name='obs')
            return Categorical(range(2)).sample(name='rv')
        else:
            return Categorical(range(3)).sample(name='rv')

    seed = 13842
    samples = 5000

    for model, expected_dist in [
        (
            model_simple,
            Categorical(range(3), probabilities=[4/7, 2/7, 1/7]),
        ),
        (
            model_branching,
            Categorical(range(3), probabilities=[
                1/6 * 1/2 + 5/6 * 1/3,
                1/6 * 1/2 + 5/6 * 1/3,
                5/6 * 1/3,
            ]),
        ),
    ]:
        print('model', model)

        dist = _distribution_from_inference(SimpleEnumeration(model).run())
        print('Enumeration', dist)
        assert dist.isclose(expected_dist)

        dist = _distribution_from_inference(LikelihoodWeighting(model, samples=samples, seed=seed).run())
        print('LikelihoodWeighting', dist)
        assert dist.isclose(expected_dist, atol=1e-1)

def test_graph_enumeration():
    def f1():
        def g():
            return flip(.4) + flip(.7) + flip(.9) + flip(.2) + flip(.51)
        return g() + g()

    def f2():
        def g(i):
            return flip() + flip()
        g = mem(g)
        i = flip()
        j = flip()
        return g(i) + g(j)

    def f3():
        i = flip(.3)
        j = flip(.72)
        condition(.9 if i + j == 1 else .3)
        return i + j

    def f4():
        # @register_call_entryexit
        def g(i):
            return flip(.61, name='a') + flip(.77, name='b')
        x = flip(.3, name='x')
        return x + g(1)

    def f5():
        # @register_call_entryexit
        def g(i):
            Bernoulli(.3).observe(i)
            return flip(.61, name='a') + flip(.77, name='b')
        x = flip(.3, name='x')
        return x + g(x)

    def f6():
        num = lambda : draw_from(range(2))
        op = lambda : '+' if flip(.5) else '*'
        def eq(d):
            if d == 0 or flip(.34):
                return num()
            else:
                return (num(), op(), eq(d - 1))
        return eq(3)

    def f7():
        return flip(0)

    def f8():
        def g(i):
            if i < 0.5:
                return 1
            else:
                return flip(i)
        x = independent_map(g, (.1, .2, .3, .4, .5, .6, .7))
        return x

    def f9():
        # @register_call_entryexit
        def g(i):
            return flip(i)
        x = g(.2)
        condition(1)
        return x

    def f10():
        def g(i):
            return flip(i)
        x = independent_map(g, (.2,))
        condition(1)
        return x

    def f11():
        # @register_call_entryexit
        def f(a, b):
            condition(a == b)
        a = flip()
        b = flip()
        f(a, b)
        return a + b

    test_models = [f1, f2, f3, f4, f5, f6, f7, f8, f9, f10, f11]

    failed = {}
    for f in test_models:
        e_res = SimpleEnumeration(f).run()
        ge_res = Enumeration(f).run()
        if not e_res.isclose(ge_res):
            failed[f] = (e_res, ge_res)

    if failed:
        raise AssertionError('\n'.join([
            "Failed models:",
            *[f"{f.__name__}:\n\t{e_res}\n\t{ge_res}" for f, (e_res, ge_res) in failed.items()]
        ]))

def test_hashing_program_states_with_list_and_dict():
    def f():
        a = []
        b = {}
        return flip()

    ps = CPSInterpreter().initial_program_state(f)
    assert id(ps.step()) != id(ps.step())
    assert hash(ps.step()) == hash(ps.step())

def test_graph_enumeration_callsite_caching():
    def model():
        @register_call_entryexit
        def f(p):
            return flip(p) + flip(p)
        x = f(.3) + f(.3) + f(.3)
        return x

    e_res = SimpleEnumeration(model).run()
    ge_res_nocache = Enumeration(model, _call_cache_size=0, _emit_call_entryexit=False).run()
    ge_res_cache = Enumeration(model, _call_cache_size=1000, _emit_call_entryexit=False).run()
    assert e_res.isclose(ge_res_nocache)
    assert e_res.isclose(ge_res_cache)

def test_graph_enumeration_callsite_caching_with_mem():
    def model():
        @mem
        def g(i):
            return .3 if flip(.6) else .63
        @register_call_entryexit
        def f():
            p = g(0)
            return flip(p) + flip(p)
        x = f() + f() + f()
        return x

    e_res = SimpleEnumeration(model).run()
    ge_res_nocache = Enumeration(model, _call_cache_size=0, _emit_call_entryexit=False).run()
    ge_res_cache = Enumeration(model, _call_cache_size=1000, _emit_call_entryexit=False).run()
    assert e_res.isclose(ge_res_nocache)
    assert e_res.isclose(ge_res_cache)

def test_graph_enumeration_callsite_caching_lru_cache():
    def model():
        @register_call_entryexit
        def f(p):
            return flip(p)
        x = f(.1) + f(.2) + f(.3) + f(.4)
        return x

    ge = Enumeration(model, _call_cache_size=2, _emit_call_entryexit=False)
    e_res = SimpleEnumeration(model).run()
    ge_res_cache = ge.run()
    assert e_res.isclose(ge_res_cache)
    assert len(ge._call_cache) == 2
    assert [args[0] for _, args, _, _ in ge._call_cache.keys()] == [.3, .4]

def test_Enumeration_call_cache_outside_function():
    # after m(1):
    #     misses = {f(-1), f(-2)}     n = 2
    #     hits = [f(-1)]    n = 1
    # after m(2):
    #     misses = {f(-1), f(-2), f(-3)}   n = 3
    #     hits = [f(-1), f(-2), f(-2)]   n = 3
    def f(i):
        return 0
    def m(i):
        return f(-i) + f(-i) + f(-(i + 1))

    enum = Enumeration(m, _call_cache_size=10, _emit_call_entryexit=True)
    enum.run(1)
    assert enum._call_cache.hits == 1
    assert enum._call_cache.misses == 2
    assert len(enum._call_cache) == 2
    enum.run(2)
    assert enum._call_cache.hits == 3
    assert enum._call_cache.misses == 3
    assert len(enum._call_cache) == 3

def test_Enumeration_call_cache_nested_function():
    # after m(1):
    #     misses = {f(-1), f(-2)}     n = 2
    #     hits = [f(-1)]    n = 1
    # after m(2):
    #     misses = {f(-1), f(-2), f(-3)}   n = 3
    #     hits = [f(-1), f(-2), f(-2)]   n = 3
    def m(i):
        def f(i):
            return 0
        return f(-i) + f(-i) + f(-(i + 1))

    enum = Enumeration(m, _call_cache_size=10, _emit_call_entryexit=True)
    enum.run(1)
    assert enum._call_cache.hits == 1
    assert enum._call_cache.misses == 2
    assert len(enum._call_cache) == 2
    enum.run(2)
    assert enum._call_cache.hits == 3
    assert enum._call_cache.misses == 3
    assert len(enum._call_cache) == 3

    def m(i):
        def f(i):
            return i + 1
        return f(f(f(i)))

    enum = Enumeration(m, _call_cache_size=10, _emit_call_entryexit=True)
    enum.run(1)
    assert enum._call_cache.hits == 0
    assert enum._call_cache.misses == 3
    assert len(enum._call_cache) == 3

def test_Enumeration_binom():
    def binom(k, p):
        ct = 0
        for i in range(k):
            ct += Bernoulli(p).sample()
        return ct

    enum = Enumeration(binom, _call_cache_size=100, _emit_call_entryexit=True)
    d = enum.run(2, .5)
    assert enum._call_cache.hits == 1
    assert enum._call_cache.misses == 6
    assert len(enum._call_cache) == 6
    assert d.isclose(Categorical.from_dict({0: 0.25, 1: 0.5, 2: 0.25}))

    simpler_keys = [
        (scope.get('i'), scope.get('ct'))
        for (_, (scope,), _, _) in enum._call_cache.keys()
    ]
    assert len(set(simpler_keys)) == 6
    binom2keys = {
        (None, 0): 1,
        (0, 0): 1,
        (0, 1): 1,
        (1, 0): 1,
        (1, 1): 1,
        (1, 2): 1,
    }
    assert collections.Counter(simpler_keys) == binom2keys

    enum = Enumeration(binom, _call_cache_size=100, _emit_call_entryexit=True)
    d = enum.run(3, .5)
    assert enum._call_cache.hits == 3
    assert enum._call_cache.misses == 10
    assert len(enum._call_cache) == 10
    assert d.isclose(Categorical.from_dict({0: 1/8, 1: 3/8, 2: 3/8, 3: 1/8}))

    simpler_keys = [
        (scope.get('i'), scope.get('ct'))
        for (_, (scope,), _, _) in enum._call_cache.keys()
    ]
    assert len(set(simpler_keys)) == 10
    assert collections.Counter(simpler_keys) == {
        **binom2keys,
        (2, 0): 1,
        (2, 1): 1,
        (2, 2): 1,
        (2, 3): 1,
    }

def test_enumerating_class_method():
    def flip():
        p = .66
        return Bernoulli(p).sample()

    class C:
        p = .66
        def flip(self):
            return Bernoulli(self.p).sample()
    c = C()
    method_res = Enumeration(C.flip).run(c)
    func_res = Enumeration(flip).run()
    assert method_res.isclose(func_res)

def test_Enumeration__set_cont_var_func():
    def f():
        return Uniform().sample()

    def model():
        p = f()
        return p

    continous_handler1 = lambda ps: [0.5]
    continous_handler2 = lambda ps: [0.25]

    enum = Enumeration(
        model,
        cont_var_func=None,
        _emit_call_entryexit=True
    )

    with pytest.raises(AssertionError) as e:
        enum.run()
    assert "Continuous sample state values must be provided for continuous distributions" in str(e.value)

    enum.set_cont_var_func(continous_handler1)
    res1 = enum.run()
    assert res1.isclose(Categorical.from_dict({0.5: 1.0}))

    enum.set_cont_var_func(continous_handler2)
    res2 = enum.run()
    assert res2.isclose(Categorical.from_dict({0.25: 1.0}))

def test_Enumeration__max_states_with_recursive_model():
    def deep_recurse(d):
        if d == 0:
            return 0
        i = Bernoulli(.9).sample(name=f'i{d}') #asymmetric otherwise its just bfs
        return i + deep_recurse(d - 1)

    n_sample_states_visited = [0]
    def cb(ps):
        if isinstance(ps, SampleState):
            n_sample_states_visited[0] += 1
        return

    enum_limit = Enumeration(
        deep_recurse,
        max_states=10,
        _emit_call_entryexit=False,
        _state_visit_callback=cb
    )
    enum_limit_res = enum_limit.run(5)
    assert n_sample_states_visited[0] == 11

    enum_full_res = Enumeration(deep_recurse).run(5)
    assert not enum_full_res.isclose(enum_limit_res)

def test_Enumeration__set_cont_var_func__with_without_entry_exit():
    def f():
        return Uniform().sample()

    def model():
        return f() + f()

    cont_func_no_entryexit_visits = [0]
    def cont_func_no_entryexit(ps: SampleState):
        cont_func_no_entryexit_visits[0] += 1
        return [.1]

    Enumeration(
        model,
        _emit_call_entryexit=False,
        cont_var_func=cont_func_no_entryexit
    ).run()

    cont_func_with_entryexit_visits = [0]
    def cont_func_with_entryexit(ps: SampleState):
        cont_func_with_entryexit_visits[0] += 1
        return [.1]

    Enumeration(
        model,
        _emit_call_entryexit=True,
        cont_var_func=cont_func_with_entryexit
    ).run()

    # both continuous variable functions should have been accessed twice since
    # they should not be cached
    assert cont_func_no_entryexit_visits[0] == 2
    assert cont_func_with_entryexit_visits[0] == 2

def test_Enumeration__set_cont_var_func__with_without_entry_exit_nested():
    # we should always be calling into g(None) and then f()
    def f():
        return Uniform().sample()
    def g(p=None):
        if p is None:
            p = f()
        return p

    def model():
        return g() + g(.3) + g(.3) + g() + g(.3) + g()

    cont_var_func_visits = [0]
    def cont_var_func(ps: SampleState):
        cont_var_func_visits[0] += 1
        return [.1]
    enum = Enumeration(model, cont_var_func=cont_var_func, _emit_call_entryexit=True)
    dist = enum.run()
    assert dist.isclose(Categorical([.1 + .3 + .3 + .1 + .3 + .1]))
    assert cont_var_func_visits[0] == 3
    assert len(enum._call_cache) == 1
    assert enum._call_cache.hits == 2
    assert enum._call_cache.misses == 2 + 1 + 0 + 2 + 0 + 2

def test_Enumeration__cont_var_func__pre_continuous_branching():
    def f(p=None):
        if p is None:
            p = Uniform().sample()
        return p

    def model():
        _ = Bernoulli(.3).sample()
        return f(.73) + f(.73) + f() + f()

    cont_var_func_visits = [0]
    def cont_var_func(ps):
        cont_var_func_visits[0] += 1
        return [.4]

    enum = Enumeration(model, cont_var_func=cont_var_func)
    dist = enum.run()
    assert dist.isclose(Categorical([.73 + .73 + .4 + .4]))
    assert cont_var_func_visits[0] == (0 + 0 + 1 + 1) + (0 + 0 + 1 + 1)
    assert len(enum._call_cache) == 1
    assert enum._call_cache.hits == (0 + 1 + 0 + 0) + (1 + 1 + 0 + 0)
    assert enum._call_cache.misses == (1 + 0 + 1 + 1) + (0 + 0 + 1 + 1)

def test_Enumeration__cont_var_func__post_continuous_branching():
    def f(p=None):
        if p is None:
            p = Uniform().sample()
        return p

    def model():
        res = f(.73) + f(.73) + f() + f()
        _ = Bernoulli(.3).sample()
        return res

    cont_var_func_visits = [0]
    def cont_var_func(ps):
        cont_var_func_visits[0] += 1
        return [.4]

    enum = Enumeration(model, cont_var_func=cont_var_func)
    dist = enum.run()
    assert dist.isclose(Categorical([.73 + .73 + .4 + .4]))
    assert cont_var_func_visits[0] == (0 + 0 + 1 + 1)
    assert len(enum._call_cache) == 1
    assert enum._call_cache.hits == (0 + 1 + 0 + 0)
    assert enum._call_cache.misses == (1 + 0 + 1 + 1)

def test_MaximumMarginalAPosteriori__fit_binomial():
    def model1():
        p = Uniform().sample()
        x = Binomial(9, p).sample()
        condition(x == 6)
        return p

    dist = MaximumMarginalAPosteriori(model1, seed=142).run()
    assert len(dist) == 1
    assert isclose(dist.sample(), 6/9, atol=1e-4)

    def model2():
        p1 = Uniform().sample()
        p2 = Uniform().sample()
        p = p1 + p2
        condition(p < 1)
        Binomial(9, p).observe(6)
        return p

    dist = MaximumMarginalAPosteriori(model2, seed=42).run()
    assert len(dist) == 1
    assert isclose(dist.sample(), 6/9, atol=1e-4)

    def model3():
        p1 = Uniform().sample()
        p2 = Uniform().sample()
        assert p1 != p2
        p = p1 if Bernoulli(.5).sample() else p2*.9
        Binomial(9, p).observe(6)
        return p

    dist = MaximumMarginalAPosteriori(model3, seed=142).run()
    assert isclose(dist.expected_value(), 6/9, atol=1e-4)

def test_MaximumMarginalAPosteriori__fit_gaussian():
    def m1(data, prior_mu=0, prior_sig=1, obs_sig=.1):
        mu = Gaussian(prior_mu, prior_sig).sample()
        [Gaussian(mu, obs_sig).observe(d) for d in data]
        return mu

    # conjugate update
    def m1_conj(data, prior_mu=0, prior_sig=1, obs_sig=.1):
        post_sig = 1/((1/prior_sig**2) + (len(data)/(obs_sig**2)))
        post_mu = post_sig*(prior_mu/prior_sig**2 + sum(data)/(obs_sig**2))
        return post_mu

    data = (1.579, .667, .234)
    # Maximum A Posteriori
    dist = MaximumMarginalAPosteriori(m1, maximum_likelihood=False, seed=142).run(data)
    assert len(dist) == 1
    assert isclose(dist.expected_value(), m1_conj(data), atol=1e-4)
    map_mu = dist.expected_value()
    exp_score = sum(Gaussian(map_mu, .1).log_probability(d) for d in data) + \
        Gaussian(0, 1).log_probability(map_mu)
    est_score = np.log(dist.marginal_likelihood)
    assert isclose(est_score, exp_score), (est_score, exp_score)

    # Maximum Likelihood
    dist_ml = MaximumMarginalAPosteriori(m1, maximum_likelihood=True, seed=142).run(data)
    assert len(dist_ml) == 1
    assert isclose(dist_ml.expected_value(), sum(data)/len(data), atol=1e-4)
    ml_mu = dist_ml.expected_value()
    exp_score = sum(Gaussian(ml_mu, .1).log_probability(d) for d in data)
    est_score = np.log(dist_ml.marginal_likelihood)
    assert isclose(est_score, exp_score), (est_score, exp_score)

def test_MaximumMarginalAPosteriori__fit_unbounded_number_of_vars():
    def f(i=0):
        p = Uniform().sample(name=f"p{i}")
        if p < .4:
            return i, p
        return f(i=i + 1)

    def m():
        i, p = f()
        Poisson(5).observe(i)
        Binomial(10, p).observe(5)
        return i, p

    # Nelder-Mead works here; Powell tends to get stuck or hang
    dist = MaximumMarginalAPosteriori(m, method="Nelder-Mead", seed=142).run()
    assert isclose(dist.expected_value(lambda x: x[1]), .4, atol=1e-4)

def test_DiscreteInferenceResult_from_values_scores():
    values = [0, 1, 2, 3]
    scores = [-0.1, -0.2, -0.3, -0.4]
    numpy_dist = DiscreteInferenceResult._from_values_scores_numpy(values, scores)
    builtin_dist = DiscreteInferenceResult._from_values_scores_builtin(values, scores)
    assert numpy_dist.isclose(builtin_dist)




"""
This module implements a continuation-passing style (CPS) interpreter for Python code,
which allows us to intercept sample and observe statements as described by
[van de Meent et al. (2018)](https://arxiv.org/abs/1809.10756).
"""

import ast
import functools
import inspect
import textwrap
import types
from types import MethodType
import dataclasses
import linecache
import contextlib
from typing import Union, TYPE_CHECKING, Tuple, Callable, TypeVar, Any

from flippy.core import ReturnState, SampleState, ObserveState, InitialState
from flippy.callentryexit import EnterCallState, ExitCallState
from flippy.distributions.base import Distribution, Element
from flippy.transforms import DesugaringTransform, \
    SetLineNumbers, CPSTransform, PythonSubsetValidator, ClosureScopeAnalysis, \
    GetLineNumber, CPSFunction, HashableCollectionTransform, Placeholder, \
    SetLineNumbers
from flippy.core import GlobalStore, ReadOnlyProxy
from flippy.funcutils import method_cache
from flippy.hashable import hashabledict, hashablelist

from flippy.types import NonCPSCallable, Method, Continuation, \
    SampleCallable, ObserveCallable, CPSCallable, VariableName

# Note if we use python 3.10+ we can use typing.ParamSpec
# so that combinators preserve the type signature of functions
R = TypeVar('R')

@dataclasses.dataclass(frozen=True)
class StackFrame:
    """
    @private
    The `StackFrame` class represents a single frame in the stack of function calls.
    It contains the source code of the function, the line number of the call,
    a unique call identifier, and the local variables at the time of the call.
    """
    func_src: str
    lineno: int
    call_id: int
    locals: dict

    def as_string(self):
        func_string, line_match = self._func_src_string_line_match()
        func_string = ['   '+r for r in func_string]
        func_string[line_match] = func_string[line_match].replace('  ', '>>', 1)
        func_string[line_match] = func_string[line_match] + f'  # {self.locals}'
        return '\n'.join(func_string)

    def _func_src_string_line_match(self):
        try:
            func_ast = ast.parse(self.func_src).body[0]
            line = GetLineNumber()(func_ast, self.lineno)
            func_string = ast.unparse(func_ast).split('\n')
            line_string = ast.unparse(line)
        except SyntaxError:
            func_string = self.func_src.split('\n')
            line_string = func_string[self.lineno]
        line_match = [i for i, l in enumerate(func_string) if line_string in l]
        assert len(line_match) == 1
        line_match = line_match[0]
        return func_string, line_match

    def _repr_html_(self):
        func_string, line_match = self._func_src_string_line_match()
        func_string = [r.replace('<', '&lt;').replace('>', '&gt;') for r in func_string]
        func_string[line_match] = '<span style="color:red;">'+func_string[line_match]+'</span>'
        func_html = '<pre>'+'\n'.join(func_string)+'</pre>'
        func_html = func_html.replace('  ', '&nbsp;&nbsp;')
        if self.locals is None:
            locals_html = '<pre>None</pre>'
            locals_keys = "<pre style='display:inline;'></pre>"
        else:
            locals_html = '<pre>'+'\n'.join([
                f'{k}: {v}'.replace('<', '&lt;').replace('>', '&gt;')
                for k, v in self.locals.items()
            ])+'</pre>'
            locals_keys = "<pre style='display:inline;'>"+', '.join(self.locals.keys())+'</pre>'
        func_head = "<pre style='display:inline;'>"+func_string[0].replace(":", "").replace("def ", "")+"</pre>"
        frame_html = [
            f"<details><summary>Locals: {locals_keys}</summary>{locals_html}</details>",
            f"<details><summary>Caller: {func_head} </summary>{func_html}</details>"
        ]
        frame_html = "<div style='cursor:default'>"+'\n'.join(frame_html)+"</div>"
        return frame_html

@dataclasses.dataclass(frozen=True)
class Stack:
    """
    @private
    The `Stack` class represents a stack of function call frames.
    Note that the stack is not actually used in execution since
    CPS-transformed programs are executed with a trampoline, but it provides
    a way to identify the current state of execution for debugging and
    addressing.
    """
    stack_frames: Tuple[StackFrame] = dataclasses.field(default_factory=tuple)

    def update(
        self,
        func_src: str,
        locals_: dict,
        call_id: int,
        lineno: int
    ) -> 'Stack':
        if isinstance(locals_, dict):
            locals_ = hashabledict({
                k: v for k, v in locals_.items()
                if (
                    (k not in ['__func_src', '_cont', '_cps', '_stack']) and
                    ('_scope_' not in k)
                )
            })
        new_stack = self.stack_frames + (StackFrame(func_src, lineno, call_id, locals_),)
        return Stack(new_stack)

    def as_string(self):
        return '\n'.join([f'Frame {i}:\n{frame.as_string()}\n' for i, frame in enumerate(self.stack_frames)])

    def __getitem__(self, key):
        return self.stack_frames[key]

    def _repr_html_(self):
        stack_html = []
        for i, frame in enumerate(self.stack_frames):
            frame_html = frame._repr_html_()
            frame_html = "<div style='margin-left: 20px;'>"+frame_html+"</div>"
            frame_html = f"<details open><summary>Frame {i}</summary>{frame_html}</details>"
            stack_html.append(frame_html)
        return '\n'.join(stack_html)

    def without_locals(self):
        return Stack(tuple([
            StackFrame(f.func_src, f.lineno, f.call_id, None) for f in self.stack_frames
        ]))

    def __len__(self):
        return len(self.stack_frames)

class CPSInterpreter:
    """
    @private
    The `CPSInterpreter` class is responsible for
    first transforming Python functions into CPS form using
    source-to-source transformations found in `flippy.transforms` and then
    executing the transformed code. The `CPSInterpreter` also handles the
    the interpretation of various callable types, including normal Python functions,
    class methods, static methods, and distribution methods and provides
    a way to interact with program execution with `ProgramState` objects.
    """

    _compile_mode = False
    _decorator_var_name = "__decorator_vars__"
    def __init__(self, _emit_call_entryexit: bool = False):
        self.subset_validator = PythonSubsetValidator()
        self.desugaring_transform = DesugaringTransform()
        self.hashable_collection_transform = HashableCollectionTransform()
        self.closure_scope_analysis = ClosureScopeAnalysis()
        self.setlines_transform = SetLineNumbers()
        self.cps_transform = CPSTransform()
        self.global_store_proxy = ReadOnlyProxy()
        self._emit_call_entryexit = _emit_call_entryexit

    def initial_program_state(self, call: Union['NonCPSCallable','CPSCallable']) -> InitialState:
        cps_call = self.interpret(
            call=call,
            stack = Stack(),
            func_src = "<root>",
            locals_ = {},
            call_id=None,
            lineno= 0
        )
        def return_continuation(value):
            return ReturnState(
                value=value,
                stack=Stack((StackFrame("<root>", 0, None, hashabledict({'__return__': value})), )),
            )
        def program_continuation(*args, **kws):
            return cps_call(
                *args,
                _cont=return_continuation,
                **kws
            )
        return InitialState(
            continuation=program_continuation,
            cps=self
        )

    def interpret(
        self,
        call : Union['NonCPSCallable', 'CPSCallable'],
        cont : 'Continuation' = None,
        stack : Stack = Stack(),
        func_src : str = None,
        locals_ : dict = None,
        call_id : int = None,
        lineno : int = None,
    ) -> 'Continuation':
        """
        This is the main entry point for interpreting CPS-transformed code.
        See `CPSTransform.visit_Call` in `transforms.py` for more details on
        how it appears in transformed code.

        `call_id` is used to uniquely identify function calls in the original code.
        `lineno` is used to track the line number of the call in the source
        code for debugging purposes.
        """

        # normal python
        if (
            isinstance(call, types.BuiltinFunctionType) or \
            isinstance(call, type) or
            (isinstance(call, MethodType) and isinstance(call.__self__, GlobalStore))
        ):
            continuation = self.interpret_builtin(call)
            return functools.partial(continuation, _cont=cont)

        # cps python
        if isinstance(call, MethodType):
            # Only instance/class methods (not static methods) follow this path
            continuation = self.interpret_object_attribute_call(call)
        else:
            continuation = self.interpret_cps(call)
        if not isinstance(stack, Stack):
            stack = Stack(stack)
        cur_stack = stack.update(func_src, locals_, call_id, lineno)
        continuation = functools.partial(continuation, _stack=cur_stack, _cont=cont)
        return continuation

    def interpret_builtin(self, call: 'NonCPSCallable') -> 'Continuation':
        def builtin_continuation(*args, _cont: 'Continuation'=lambda val: val, **kws):
            return lambda : _cont(call(*args, **kws))
        return builtin_continuation

    def interpret_object_attribute_call(self, call: 'NonCPSCallable') -> 'Continuation':
        if isinstance(call.__self__, Distribution):
            if call.__name__ == "sample":
                return self.interpret_sample(call, fit=False)
            elif call.__name__ == "observe":
                return self.interpret_observe(call)
            elif call.__name__ == "observe_all":
                return self.interpret_method(call) #see Distribution.observe_all
            elif call.__name__ == "fit": #fit is an alternative interface to sample
                return self.interpret_sample(call, fit=True)
            else:
                # other than sample and observe, we interpret Distribution methods as deterministic
                return self.interpret_method_deterministically(call)
        else:
            return self.interpret_method(call)

    def interpret_method(self, call: "NonCPSCallable") -> 'Continuation':
        # For instance and class methods, we transform and compile the
        # function body like normal but need to ensure the right first
        # argument is passed in.
        # This handles both class and instance methods since
        # call.__self__ refers to the the instance if a method,
        # or the class if a classmethod
        continuation = self.interpret_cps(call.__func__)
        def method_continuation(*args, _cont: 'Continuation'=None, _stack: 'Stack'=None, **kws):
            return lambda : continuation(call.__self__, *args, _cont=_cont, _stack=_stack, **kws)
        return method_continuation

    @method_cache
    def interpret_cps(
        self,
        call : Union['NonCPSCallable', 'CPSCallable']
    ) -> 'Continuation':
        if CPSTransform.is_transformed(call):
            continuation = self.interpret_transformed(call)
        else:
            continuation = self.interpret_generic(call)
        return continuation

    def interpret_transformed(self, call : CPSFunction) -> 'Continuation':
        if (
            (not self._emit_call_entryexit) or \
            (not getattr(call, "_emit_call_entryexit", True)) # hook for debugging
        ):
            return self.interpret_transformed_only(call)
        else:
            return self.interpret_transformed_and_emit_entryexit(call)

    def interpret_transformed_only(self, call : CPSFunction) -> 'Continuation':
        def generic_continuation(*args, _cont: 'Continuation'=None, _stack: 'Stack'=None, **kws):
            return call(*args, **kws, _cps=self, _stack=_stack, _cont=_cont)
        return generic_continuation

    def interpret_transformed_and_emit_entryexit(self, call : CPSFunction) -> 'Continuation':
        def continuation(*args, _cont: 'Continuation'=None, _stack: 'Stack'=None, **kws):
            kws = hashabledict(kws)
            params = dict(f=call, args=args, kwargs=kws, cps=self, stack=_stack)
            def process_controller_instructions(
                run_func: bool = True,
                cached_res: Any = None
            ):
                if run_func:
                    def process_func_result(func_res):
                        return ExitCallState(
                            **params,
                            value=func_res,
                            continuation=lambda : _cont(func_res),
                        )
                    return call(*args, _stack=_stack, _cont=process_func_result, _cps=self, **kws)
                else:
                    return ExitCallState(
                        **params,
                        value=cached_res,
                        continuation=lambda : _cont(cached_res),
                    )
            return EnterCallState(
                **params,
                continuation=process_controller_instructions,
            )
        return continuation

    def interpret_sample(self, call: 'SampleCallable[Element]', fit: bool) -> 'Continuation':
        def sample_continuation(
            _cont: 'Continuation'=None,
            _stack: 'Stack'=None,
            name: 'VariableName'=None,
            initial_value=None,
        ):
            return SampleState(
                continuation=_cont,
                distribution=call.__self__,
                name=name,
                stack=_stack,
                cps=self,
                initial_value=initial_value,
                fit=fit
            )
        return sample_continuation

    def interpret_observe(self, call: 'ObserveCallable[Element]') -> 'Continuation':
        def observe_continuation(
                value: 'Element',
                _cont: 'Continuation'=None,
                _stack: 'Stack'=None,
                name: 'VariableName'=None,
                **kws
            ):
            return ObserveState(
                continuation=lambda : _cont(None),
                distribution=call.__self__,
                value=value,
                name=name,
                stack=_stack,
                cps=self,
            )
        return observe_continuation

    def interpret_method_deterministically(self, call: Method) -> 'Continuation':
        self = call.__self__
        def method_continuation(*args, _cont: 'Continuation'=lambda val: val, _stack=None, **kws):
            return lambda : _cont(call.__func__(self, *args, **kws))
        return method_continuation

    def interpret_generic(self, call: 'NonCPSCallable') -> 'Continuation':
        trans_func = self.non_cps_callable_to_cps_callable(call)
        return self.interpret_transformed(trans_func)

    def non_cps_callable_to_cps_callable(self, call: 'NonCPSCallable') -> CPSFunction:
        assert not CPSTransform.is_transformed(call), "Callable already transformed"
        if self.is_lambda_function(call):
            raise ValueError(f'Cannot interpret lambda expressions as the outermost function. As an alternative, define as a named function.')
        call_name = self.generate_unique_method_name(call)
        code = self.transform_from_func(call, call_name)
        return self.compile_cps_transformed_code_to_function(code, call)

    def compile_cps_transformed_code_to_function(
        self,
        code: ast.AST,
        call: 'NonCPSCallable',
    ) -> CPSFunction:
        # To handle closures, we initialize functions in a temporary
        # function that unpacks the closure and returns the transformed function.

        closure = self.get_closure(call)
        closure = {
            **closure,
            CPSTransform.cps_interpreter_name: self,
            "global_store": self.global_store_proxy,
            CPSFunction.__name__: CPSFunction,
            hashabledict.__name__: hashabledict,
            hashablelist.__name__: hashablelist,
        }
        unpack = '\n'.join([
            f'{k} = closure["{k}"]'
            for k in closure.keys()
        ])
        unpacker_code = Placeholder.fill(ast.parse(textwrap.dedent(f'''
            def __closure_unpacker__(closure):
                {Placeholder.new('unpack')}
                {Placeholder.new('source')}
                return {code.body[0].name}
        ''')), dict(unpack=ast.parse(unpack).body, source=code))
        unpacker_code = SetLineNumbers()(unpacker_code)
        compiled_code = self.compile(
            f'{call.__name__}_{hex(id(call)).removeprefix("0x")}.py',
            unpacker_code
        )

        context = call.__globals__
        try:
            assert CPSInterpreter._compile_mode is False
            CPSInterpreter._compile_mode = True
            exec(compiled_code, context)
            # Call the closure unpacker to get the transformed function
            trans_func = context["__closure_unpacker__"](closure)
        except SyntaxError as err :
            raise err
        finally:
            CPSInterpreter._compile_mode = False

        # If the transformed function is a classmethod or staticmethod, we need to
        # extract the underlying function.
        if isinstance(trans_func, (classmethod, staticmethod)):
            trans_func = trans_func.__func__
        return trans_func

    def generate_unique_method_name(self, call: 'NonCPSCallable') -> Union[str, None]:
        # this function generates a name that won't override a non-method
        # check if the qualified name indicates it's a class method
        # see https://peps.python.org/pep-3155/
        qualname_suffix = call.__qualname__.split("<locals>.")[-1]
        defined_in_class = len(qualname_suffix.split('.')) > 1
        if defined_in_class:
            return "__"+qualname_suffix.replace('.', '_')
        return None

    def transform_from_func(self, call: 'NonCPSCallable', call_name: str = None) -> ast.AST:
        source = inspect.getsource(call)
        source = textwrap.dedent(source)
        trans_node = ast.parse(source)
        if call_name is not None:
            self.rename_class_method_in_source(trans_node, call_name)
        self.subset_validator(trans_node, source)
        return self.transform(trans_node, source)

    def rename_class_method_in_source(self, node: ast.Module, name: str):
        assert len(node.body) == 1 and isinstance(node.body[0], ast.FunctionDef), \
            "We assume there's only a single function definition in the source"
        node.body[0].name = name

    def transform(self, trans_node: ast.AST, source: str) -> ast.AST:
        self.closure_scope_analysis(trans_node, source)
        trans_node = self.desugaring_transform(trans_node)
        trans_node = self.setlines_transform(trans_node)
        trans_node = self.cps_transform(trans_node)
        trans_node = self.hashable_collection_transform(trans_node)
        # print(ast.unparse(trans_node))
        return trans_node

    def compile(self, filename: str, node: ast.AST) -> types.CodeType:
        source = ast.unparse(node)
        # In order to get stack traces that reference compiled code, we follow the scheme IPython does
        # in CachingCompiler.cache, by adding an entry to Python's linecache.
        # https://github.com/ipython/ipython/blob/47abb68a/IPython/core/compilerop.py#L134-L178
        linecache.cache[filename] = (
            len(source),
            None,
            [line + "\n" for line in source.splitlines()],
            filename,
        )
        return compile(source, filename, 'exec')

    @staticmethod
    def get_closure(func: 'NonCPSCallable') -> dict:
        if getattr(func, "__closure__", None) is not None:
            closure_keys = func.__code__.co_freevars
            closure_values = [cell.cell_contents for cell in func.__closure__]
            return dict(zip(closure_keys, closure_values))
        else:
            return {}

    @classmethod
    def is_lambda_function(cls, func):
        return isinstance(func, types.LambdaType) and func.__name__ == "<lambda>"

    @contextlib.contextmanager
    def set_global_store(self, store : GlobalStore):
        assert self.global_store_proxy.proxied is None, 'Nested update of global store not supported.'
        try:
            self.global_store_proxy.proxied = store
            yield
        finally:
            self.global_store_proxy.proxied = None


class DescriptorMixIn:
    """
    @private
    A mixin class that provides a descriptor interface for a callable object.

    In Python, a descriptor is an object attribute with "binding behavior",
    meaning that its attribute access is overridden by methods in the descriptor protocol.
    For more details, see [here](https://docs.python.org/3/howto/descriptor.html).

    For our purposes, this mixin class implements a descriptor interface for
    callables that wrap class methods, static methods, and normal methods that is compatible with
    the CPSInterpreter.
    """
    def __init__(self, wrapped_func):
        if inspect.ismethod(wrapped_func):
            raise ValueError("Cannot wrap a method outside the class namespace")
        self.is_classmethod = isinstance(wrapped_func, classmethod)
        self.is_staticmethod = isinstance(wrapped_func, staticmethod)

    def __call__(self, *args, _cont=None, _cps=None, _stack=None, **kws):
        raise NotImplementedError

    def __get__(self, obj, objtype=None):
        # if we wrapped a class method, we return a partial function with cls
        if self.is_classmethod:
            objtype = objtype if objtype is not None else type(obj)
            partial_call = functools.partial(self.__call__, objtype)
            setattr(partial_call, CPSTransform.is_transformed_property, True)
            return partial_call

        # if we wrapped a static method, we return the function itself
        if self.is_staticmethod:
            return self

        # if we wrapped a normal method, we return the function itself or
        # a partial function with obj
        if obj is None: # this means the function is not bound to an instance
            return self
        partial_call = functools.partial(self.__call__, obj)
        setattr(partial_call, CPSTransform.is_transformed_property, True)
        return partial_call

class KeepDeterministicCallable(DescriptorMixIn):
    '''
    @private
    '''
    def __init__(self, func):
        DescriptorMixIn.__init__(self, func)
        self.wrapped_func = func
        if isinstance(func, (classmethod, staticmethod)):
            self.wrapped_func = func.__func__
        functools.update_wrapper(self, func)
        # This ensures that no subsequent compilation happens.
        setattr(self, CPSTransform.is_transformed_property, True)

    def __call__(self, *args, _cont=None, _cps=None, _stack=None, **kws):
        rv = self.wrapped_func(*args, **kws)
        if _cont is None:
            return rv
        else:
            return lambda : _cont(rv)


def keep_deterministic(fn: Callable[..., R]) -> Callable[..., R]:
    '''
    Decorator to interpret a function as deterministic Python.
    Any random sampling in the function will not be targeted for inference.
    Any conditioning will not be incorporated into the inference process.

    This is helpful if the transform slows a function down, if a
    deterministic library is being called, or if a distribution is being
    directly computed.
    '''
    return KeepDeterministicCallable(fn)

def cps_transform_safe_decorator(dec: Callable) -> Callable:
    """
    A higher-order function that wraps a decorator so that it works with functions
    that are CPS transformed or will be CPS transformed.

    Functions are typically evaluated twice in the course of being CPS transformed.
    First, when they are initially declared in Python.
    Second, when being evaluated after being transformed.
    So, a function's decorators would typically be executed twice. However, this decorator
    can be used to mark a decorator so that it will not be executed after being CPS transformed.
    The resulting function will then be the transformed version of decorated function, without
    any further decoration after transformation.
    """
    @keep_deterministic
    def wrapped_decorator(fn=None, *args, **kws):
        if fn is None:
            return functools.partial(wrapped_decorator, *args, **kws)

        # This code mainly supports two cases:
        # 1. We need to CPS transform and wrap a function (e.g., when decorating in the module scope)
        # 2. We only need to wrap a function (e.g., when decorating in a nested function scope)

        # Case 1 is confusing because the transformation and wrapping occur in two different
        # calls to the current code. Specifically, we visit the current
        # code first from the module scope and then transform/wrap it using a CPSInterpreter
        # object. The CPSInterpreter object transforms the source code and then executes the current
        # code a second time on the transformed function (in an "exec" statement). This second
        # call to the current code does not need to transform the function, and this is indicated
        # by a _compile_mode flag associated with the CPSInterpreter class.

        # Case 2 is simpler because we only need to wrap the function. This occurs when the function
        # is nested inside another function that has already been CPS transformed.
        if CPSInterpreter._compile_mode:
            return fn
        wrapped_fn = dec(fn, *args, **kws)
        return wrapped_fn
    wrapped_decorator = functools.wraps(dec)(wrapped_decorator)
    return wrapped_decorator



"""
Several inference algorithms are available depending on the use case.
For discrete models with small state spaces, exact inference can be performed
using enumeration. For larger discrete models and models with continuous variables,
approximate inference can be performed using sampling-based methods.
"""

from flippy.inference.inference import InferenceAlgorithm
from flippy.inference.simpleenumeration import SimpleEnumeration
from flippy.inference.enumeration import Enumeration
from flippy.inference.sample_prior import SamplePrior
from flippy.inference.likelihood_weighting import LikelihoodWeighting
from flippy.inference.mcmc.metropolis_hastings import MetropolisHastings
from flippy.inference.max_marg_post import MaximumMarginalAPosteriori
from flippy.distributions import Categorical

__all__ = [
    "SimpleEnumeration",
    "Enumeration",
    "SamplePrior",
    "LikelihoodWeighting",
    "MetropolisHastings",
    "MaximumMarginalAPosteriori",
]

def _distribution_from_inference(dist):
    ele, probs = zip(*dist.items())
    return Categorical(ele, probabilities=probs)
